;###############################################################################
/* Selector for LilBuilder

*/
;###############################################################################

Selector_GuiID(){                       ;selectors own GuiID
  Return 1
}

Selector_HWND(NewHWND=""){              ;selectors own gui HWND
  static HWND
  If NewHWND
     HWND = %NewHWND%
  Return %HWND%
}
                                        
Selector_Create() {                     ;create selector gui
   local Selector_GuiID
   Selector_GuiID := Selector_GuiID()
   Gui, %Selector_GuiID%:+LastFound -Caption +Resize +ToolWindow +LabelSelector_ ;AlwaysOnTop
   Gui, %Selector_GuiID%:Color, FFFF
   WinSet, Transparent, 100
   Selector_HWND(WinExist())
}

Selector_SelCtrlHWND(NewHWND=""){       ;currently selected control
  static HWND
  If NewHWND = No
     HWND =
  Else If NewHWND
     HWND = %NewHWND%
  Return %HWND%
}

Selector_SelGuiHWND(NewHWND=""){        ;currectly "active" (selected) user gui
  static HWND
  If NewHWND = No                       ;there should normally always be one, toolbox relies on it
     HWND =
  Else If NewHWND
     HWND = %NewHWND%
  Return %HWND%
}

Selector_SelMultiHWND(NewHWND=""){      ;list of selected controls in multi selection
  static HWND
  If NewHWND = No
     HWND =
  Else If NewHWND
     HWND .= "|" NewHWND
  Return %HWND%
}

;store and retrieve position and size of selector when in multiselection
Selector_MultiPosSize(ByRef X, ByRef Y, ByRef W, ByRef H){
  static sX, sY, sW, sH
  If X
    sX := X, sY := Y, sW := W, sH := H,
  Else
    X := sX, Y := sY, W := sW, H := sH,
}

;set position and size of selector gui
Selector_Position(X, Y, W, H){
  Selector_HWND := Selector_HWND()
  Selector_GuiID := Selector_GuiID()
  WinMove, ahk_id %Selector_HWND% , ,x, y, w, h
  Gui, %Selector_GuiID%:Show
  WinActivate, ahk_id %Selector_HWND%
}

;selector gui gets resized
Selector_Size:
  Selector_Size()
return
Selector_Size(State=""){
  If State {         ;when called with state
    GuiID := Selector_GuiID()
    Gui, %GuiID%:%State%Resize     ;Change resize feature
    Return         
  }
    
  If Selector_SelCtrlHWND() {                  ;when single selection  
    If (GRID() AND !GetKeyState("LControl")){       ;when Grid and NOT LCtrl down
      Selector_UpdateCtrlwGrid()                         ;update control in grid
      SetTimer, Selector_UpdateItself, 100               ;update selector to new control position when resize finished
    }
    else Selector_UpdateCtrl()                     ;otherwise move the control directly to follow the selector
  }Else                                            ;when multi selection
    SetTimer, Selector_UpdateMultiSelOnResize, 10  ;update controls according to new selector pos/size when resize finished
}

;update controls according to new selector pos/size when resize finished
Selector_UpdateMultiSelOnResize:
  If !GetKeyState("LButton"){                           ;wait for LButton release, then resize finsihed
      SetTimer, Selector_UpdateMultiSelOnResize, Off         ;disable timer 
      Selector_UpdateCtrl()                                  ;update controls to new selector position/size 
      Grid_Draw()                                            ;redraw grid 
  }
Return

;add a control to a multi selection
; called when:
;   - a single control gets selected with LCtrl down
Selector_AddCtrlToSelection(CtrlHwnd){
  Selector_Size("-")                                 ;disable resize function of selector
  If (SingelSelCrtlHWND := Selector_SelCtrlHWND()) { ;when a single control was selected previously
    Selector_SelMultiHWND(SingelSelCrtlHWND)              ;add control from single selection to multiselection as first control
    Selector_SelCtrlHWND("No")                            ;remove control from single selection
  }
  Selector_SelMultiHWND(CtrlHwnd)                    ;add current control to multiselection
  Selector_Set2MultiSel()                            ;change pos/size of selector to cover all controls 
  Sleep, 50
  Selector_Size("+")                                 ;enable resize function of selector
}

;set selector to the dimensions of a multiselection
; called when:
;   - a single control gets add to multiselection
;   - multiple controls get selected with frame selection
Selector_Set2MultiSel(){
  MultiSelCtrlHWNDs := Selector_SelMultiHWND()
  maxx := 0 ,            maxy := 0                   ;set extreme start coordinates for surrounding box
  minx := A_ScreenWidth, miny := A_ScreenHeight   
  Loop, Parse, MultiSelCtrlHWNDs, |                  ;go through all controls in selection
    {
      If !A_LoopField                                ;skip first empty field, since list starts with |
        Continue
      ControlGetPos, x,y,w,h,, ahk_id %A_LoopField%  ;get control pos/size
      maxx := maxx > x + w ? maxx : x + w            ;calculate surrounding box coordinates
      maxy := maxy > y + h ? maxy : y + h
      minx := minx < x ? minx : x
      miny := miny < y ? miny : y
    }
  sW := maxx - minx , sH := maxy - miny              ;calculate size of surrounding box
  WinGetPos, ux, uy,,, % "ahk_id " Selector_SelGuiHWND()   ;get position of user gui
  sX := minx + ux , sY := miny + uy                        ;calculate global selector coordinates 
  Selector_Position(sX, sY, sW, sH)                        ;set new pos/size for selector
  Selector_MultiPosSize(sX, sY, sW, sH)                    ;store new pos/size
}

;start draging a frame for selection
;called when:
; - user gui got clicked with LButton without selecting a control and LButton is held down for a period of xxx ms
; Selector_StartFrameDrag:
;   If !GetKeyState("LButton")          ;if LButton got release before period finished, don't do anything
;     Return
;   Messages_SetHandlers( False )            ;disable Message handlers, e.g. to prevent accidentaly OnMove, which can happen when moving mouse fast  
;   Selector_Size("-")                       ;disable resize function of selector
;   Selector_DrawFrame()                     ;draw frame (selector gui) on mouse move
;   Selector_SelectByBox()                   ;select controls based on frame pos/size
;   If Selector_SelCtrlHWND()                ;update selector to new single selection
;     Selector_Set2Ctrl()                        
;   Else If Selector_SelMultiHWND()          ;update selector to new multi selection 
;     Selector_Set2MultiSel()                   
;   Messages_SetHandlers( True )             ;enable Message handlers again
;   Selector_Size("+")                       ;enable resize function of selector
; Return

Selector_IsDrawingFrame(NewState=""){
  Static State
  If NewState is not space
     State = %NewState%
  Return %State%
}
Selector_DrawFrame(X="", Y="", Action=""){
  static sX, sY
  If X
    sX := X, sY := Y
  MouseGetPos, eX, eY
  X := sX > eX ? eX : sX, Y := sY > eY ? eY : sY  ;calculate frame position
  W := Abs(sX - eX), H := Abs(sY - eY)            ;calculate size
  If (X AND Y AND W AND H)
    Selector_Position(X, Y, W, H)
  If Action = End
    Selector_MultiPosSize(x,y,w,h)
}

;select controls which are inside a box
; called when:
;   - frame was dragged to select multi controls
Selector_SelectByBox(){
  GuiHWND := Selector_SelGuiHWND()                          ;which Gui
  WinGetPos, ux, uy,,, % "ahk_id " Selector_SelGuiHWND()   ;get pos of user gui
  Selector_MultiPosSize(BoxX, BoxY, BoxW, BoxH)             ;get box dimentions
  BoxX -= ux, BoxY -= uy                                    ;relative position
  BoxW += BoxX, BoxH += BoxY
  
  ListOfAllControls := Gui_GetListOfAllControls(GuiHWND)

  Loop, Parse, ListOfAllControls, |                         ;go through all controls on gui
    {                                           
      ControlGetPos, cx,cy,cw,ch,, ahk_id %A_LoopField%     ;get current controls pos
      If ( cx >= BoxX AND cy >= BoxY AND cx + cw <= BoxW AND cy + ch <= BoxH )
        ControlsInBox .= "|" . A_LoopField   
    }
  StringTrimLeft, ControlsInBox, ControlsInBox, 1
  If InStr(ControlsInBox,"|")
      Selector_SelMultiHWND(ControlsInBox)                  ;set multiselection
  Else If ControlsInBox
      Selector_SelCtrlHWND(ControlsInBox)                   ;set single selection
  Else
      Return 0
}

;update selector to control dimentions
; called when: 
;   - after selector resize with GRID in single selection mode
;   - after selector move with GRID in single selection mode
Selector_UpdateItself:
  If !GetKeyState("LButton"){
      SetTimer, Selector_UpdateItself, Off
      Selector_Set2Ctrl()                    ;update selector to control dimentions
      Grid_Draw()                            ;Redraw grid
  }
Return

;set selector to the dimensions of a control
; called when:
;   - selector needs to update itself after selector got resized/moved with GRID in single selection mode
;   - after a new control gets added via toolbox
;   - a single control gets selected with a click
;   - a single control gets selected with frame selection
Selector_Set2Ctrl(){
   Selector_Size("-")                                           ;disable resize function of selector
   WinGetPos, ux, uy,,, % "ahk_id " Selector_SelGuiHWND()       ;get pos of user gui
   ControlGetPos, sx,sy,sw,sh,, % "ahk_id " Selector_SelCtrlHWND()  ;get control pos/size
   Selector_Position(sx + ux, sy + uy, sw, sh)                  ;show selector at pos/size of control 
   Sleep, 50     ;needed, otherwise Selector_Size() doesn't get False state
   Selector_Size("+")                                          ;enable resize function of selector
}

; Update control pos/size to match selector's size and position
; called when:
;   - selector resizes in single selection
;   - after selector resize in multi selection
;   - after mouse move with LButton down on selector  
Selector_UpdateCtrl(){
  WinGetPos, ux, uy,,, % "ahk_id " Selector_SelGuiHWND()        ;get pos of user gui
  WinGetPos, sx,sy,sw,sh, % "ahk_id " Selector_HWND()           ;get current pos/size of selector
  If (CtrlHWND := Selector_SelCtrlHWND()){                      ;when single selection
    ControlMove,, sx-ux, sy-uy, sw, sh, ahk_id %CtrlHWND%            ;move control to selector pos/size
  }
  Else If (ListCtrlHWND := Selector_SelMultiHWND()){            ;when multi selection
    Selector_MultiPosSize(dx, dy, dw, dh)                            ;get selectors previous pso/size
    fx := sw / dw , fy := sh / dh                                    ;calculate factor for x/y
    dx := sx - dx + (fx - 1) * (ux - dx) , dy := sy - dy + (fy - 1) * (uy - dy) ;calculate offset for x/y 
    
    ;??? doesn't work well on resize: controls width/height are not adjusted at all
    ;changing controls position based on change in selector width/height doesn't work well too => truncation error 
    ;don't know how to do it better                                         
    ;need something like Anchor, I guess                                    
    
    Loop, Parse, ListCtrlHWND, |                                     ;go through all controls in selection
      {                                           
        If !A_LoopField                                                      ;skip first empty field, since list starts with |
          Continue
        ControlGetPos, cx,cy,,,, ahk_id %A_LoopField%                        ;get current controls pos
        ControlMove,, cx * fx + dx, cy * fy  + dy, , , ahk_id %A_LoopField%  ;move control to new pos
      }
    Selector_MultiPosSize(sx,sy,sw,sh)                               ;store new selector pos/size
  }
  SetTimer, Gui_Refresh, -100                                        ;refresh gui after a timeout
}

;move selector by increments with hotkeys up/down/left/right
Selector_MoveIncrement:
  Selector_MoveIncrement(A_ThisHotkey)
  Selector_UpdateCtrl()                                         ;update control(s)
  Grid_Draw()                                                   ;redraw grid 
Return
Selector_MoveIncrement(Direction){
  WinGetPos, sx,sy,sw,sh, % "ahk_id " Selector_HWND()           ;get current pos/size of selector
  If Direction = Up                                             ;increment position
      sy--  
  Else If Direction = Down
      sy++
  Else If Direction = Left
      sx--
  Else If Direction = Right
      sx++
  Selector_Position(sX, sY, sW, sH)                             ;position selector
}

; Update single control pos/size to match selector's size and position when GRID is used
; called when:
;   - selector resizes
;   - after mouse move with LButton down on selector with "PositionOnly"
Selector_UpdateCtrlwGrid(Type="") {
  WinGetPos, ux, uy, , , % "ahk_id " Selector_SelGuiHWND()        ;get pos of user gui
  WinGetPos, cx, cy ,cw, ch, % "ahk_id " Selector_HWND()          ;get current pos/size of selector

  ;window frame height/width of a regular gui
  SysGet, SM_CXSIZEFRAME, 32   ;width of frame
  SysGet, SM_CYSIZEFRAME, 33   ;height of frame
  SysGet, SM_CYCAPTION, 4      ;height of caption, the whole title height is: SM_CYCAPTION + SM_CYSIZEFRAME

  ;window frame height/width of a toolwindow gui
  SysGet, SM_CXFIXEDFRAME, 7   ;width of frame
  SysGet, SM_CYFIXEDFRAME, 8   ;height of frame
  SysGet, SM_CYSMCAPTION, 51   ;height of caption, the whole title height is: SM_CYSMCAPTION + SM_CYFIXEDFRAME

  cx -= ux, cy -= uy                                              ;calculate relative position
  GRID := Grid()                                                  ;get GRID size
  dx := Mod(cx - SM_CXSIZEFRAME, GRID)                            ;calculate x/y with snap to grid
  dy := Mod(cy - SM_CYSIZEFRAME - SM_CYCAPTION, GRID)             
  dx -= (dx ? GRID : 0),cx -= dx, dy -= (dy ? GRID : 0),cy -= dy
  If !(Type = "PositionOnly"){                                    ;when also w/h should be adjusted (on selector resize)
    dw := Mod(cw, GRID), dh := Mod(ch, GRID)                         ;calculate w/h with snap to grid 
    cw -= (cw > dw ? dw : 0), ch -= (ch > dh ? dh : 0)
  }
  ControlMove,,cx,cy,cw,ch, % "ahk_id " Selector_SelCtrlHWND()    ;move control to new pos
}

;snap selector position to grid
Selector_Pos2Grid(){
  WinGetPos, ux, uy, , , % "ahk_id " Selector_SelGuiHWND()        ;get pos of user gui
  WinGetPos, sx, sy ,sw, sh, % "ahk_id " Selector_HWND()          ;get current pos/size of selector

  ;window frame height/width of a regular gui
  SysGet, SM_CXSIZEFRAME, 32   ;width of frame
  SysGet, SM_CYSIZEFRAME, 33   ;height of frame
  SysGet, SM_CYCAPTION, 4      ;height of caption, the whole title height is: SM_CYCAPTION + SM_CYSIZEFRAME

  ;window frame height/width of a toolwindow gui
  SysGet, SM_CXFIXEDFRAME, 7   ;width of frame
  SysGet, SM_CYFIXEDFRAME, 8   ;height of frame
  SysGet, SM_CYSMCAPTION, 51   ;height of caption, the whole title height is: SM_CYSMCAPTION + SM_CYFIXEDFRAME

  rx := sx - ux - SM_CXSIZEFRAME                                   ;calculate relative position
  ry := sy - uy - SM_CYCAPTION - SM_CYSIZEFRAME                                  
  GRID := Grid()                                                   ;get GRID size
  dx := Mod(rx, GRID), dy := Mod(ry, GRID)                         ;calculate delta x/y for snap to grid
  HalfGrid := Grid / 2                                             ;snap to closes grid
  dx -= (dx > HalfGrid ? GRID : 0), dy -= (dy > HalfGrid ? GRID : 0)                    
  Selector_Position(sX - dx, sY - dy, sW, sH)                      ;set selector to new position
}

;context sensitive hotkey to delete selection got pressed
Selector_OnDelete:
  Selector_OnDelete()                                 ;delete selected control(s)
  WinActivate, % "ahk_id " Selector_SelGuiHWND()      ;hide selector (move behind) by activating selected gui
  Grid_Draw()                                         ;redraw grid when control(s) deleted
return

;delete selected controls 
Selector_OnDelete(){
  If (CtrlHWND := Selector_SelCtrlHWND()){            ;single selection
    DllCall("DestroyWindow", "uint", CtrlHWND)             ;destroy control on gui
    Gui_RemoveControl(CtrlHWND)                            ;remove control from gui or tab list
    Control(CtrlHWND,"Remove")                             ;remove controls ditionary
  }
  Else If (ListCtrlHWND := Selector_SelMultiHWND()){  ;multi selection
    Loop, Parse, ListCtrlHWND, |                           ;go through selected controls
      {
        If !A_LoopField                                        ;skip first empty field, since list starts with |
          Continue
        DllCall("DestroyWindow", "uint", A_LoopField)          ;destroy control on gui
        Gui_RemoveControl(A_LoopField)                ;remove control from gui or tab list
        Control(A_LoopField,"Remove")                          ;remove controls ditionary
      }
  }
}

Selector_Hide() {    ;??? not used
   global
   WinHide, % "ahk_id " Selector_HWND()
   Grid_Draw()
}
